/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2007 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */

/*
 * VAhCoiFS K70pj
 */

#include <kernel.h>
#include <sil.h>
#include "target_serial.h"
#include "target_syssvc.h"

/*
 * WX^ݒl
 */
#define PORT2SIOPID(x)	((x) + 1)
#define INDEX_PORT(x)	((x) - 1)
#define GET_SIOPCB(x)	(&siopcb_table[INDEX_PORT(x)])

/*
 * UARTWX^`
 */
#define UART_BDH(x)		(x)
#define UART_BDL(x)		(x + 0x01)
#define UART_C1(x)		(x + 0x02)
#define UART_C2(x)		(x + 0x03)
#define UART_S1(x)		(x + 0x04)
#define UART_C3(x)		(x + 0x06)
#define UART_D(x)		(x + 0x07)
#define UART_C4(x)		(x + 0x0A)

#define S1_TDRE			(0x80)
#define S1_TC			(0x40)
#define S1_RDRF			(0x20)
#define S1_OR			(0x08)
#define S1_FE			(0x02)
#define S1_PE			(0x01)
#define C2_TIE			(0x80)
#define C2_TCIE			(0x40)
#define C2_RIE			(0x20)
#define C2_TE			(0x08)
#define C2_RE			(0x04)
#define C3_ORIE			(0x08)
#define C3_NEIE			(0x04)
#define C3_FEIE			(0x02)
#define C3_PEIE			(0x01)

/*
 * VAI/O|[gǗubNGA
 */
SIOPCB siopcb_table[TNUM_PORT];

static const uint32_t sioreg_table[TNUM_PORT] = {
	UART2_BASE,
};

#pragma inline
bool_t  sio_putready(SIOPCB* siopcb)
{
	return (sil_reb_mem((void*)UART_S1(siopcb->reg)) & S1_TC) != 0;
}

#pragma inline
bool_t sio_getready(SIOPCB* siopcb)
{
	return (sil_reb_mem((void*)UART_S1(siopcb->reg)) & S1_RDRF) != 0;
}

/*
 *  ^[Qbg̃VA
 */
void target_uart_init(ID siopid)
{
	uint32_t tmp, brfa;
	uint32_t reg = sioreg_table[INDEX_PORT(siopid)];
	uint32_t src_clock;

	/* UART̖ */
	sil_andb((void*)UART_C2(reg), ~(C2_TE|C2_RE));

	/* 1STOP BIT, 1START BIT, 8DATA bits, ParityȂ */
	sil_wrb_mem((void*)UART_C1(reg), 0);

	/* fck=60MHz */
	src_clock = BUS_CLOCK;

	tmp = src_clock / (BPS_SETTING * 16);
	sil_wrb_mem( (void*)UART2_BDH, (sil_reb_mem( (void*)UART2_BDH ) & ~(0x1F)) | ((tmp & 0x1F00) >> 8) );
	sil_wrb_mem( (void*)UART2_BDL, (char_t)(tmp & 0xFF) );

	brfa = (((src_clock / 1000 * 32000) / (BPS_SETTING * 16)) - (tmp * 32) );
	sil_wrb_mem( (void*)UART2_C4, (sil_reb_mem( (void*)UART2_C4 ) & ~(0x1F)) | (brfa & 0x1F) );

	/* G[݂̗L */
	sil_orb((void*)UART_C3(reg), (C3_ORIE|C3_NEIE|C3_FEIE|C3_PEIE));

	/* M̗L */
	sil_orb( (void*)UART2_C2, (C2_RIE|C2_TE|C2_RE) );
}

/*
 *  ^[Qbg̃VAI
 */
void target_uart_term(ID siopid)
{
	uint32_t reg = sioreg_table[INDEX_PORT(siopid)];

	/* UART̖ */
	sil_andb((void*)UART_C2(reg), ~(C2_TE|C2_RE));
}

/*
 *  SIO
 */
void sio_initialize(intptr_t exinf)
{
	int i;

	for (i = 0; i < TNUM_PORT; i++) {
		siopcb_table[i].port = i;
		siopcb_table[i].reg = sioreg_table[i];
		siopcb_table[i].exinf = 0;
	}
}

/*
 *  VAI[v
 */
SIOPCB *sio_opn_por(ID siopid, intptr_t exinf)
{
	SIOPCB* siopcb;

	if (siopid > TNUM_PORT) {
		return NULL;
	}

	siopcb = GET_SIOPCB(siopid);
	siopcb->exinf = exinf;

	target_uart_init(siopid);

	return siopcb;
}

/*
 *  VAN[Y
 */
void sio_cls_por(SIOPCB *p_siopcb)
{
	target_uart_term(PORT2SIOPID(p_siopcb->port));
}

/*
 *  ݃nh
 */
void sio_isr(intptr_t exinf)
{
	SIOPCB* siopcb = GET_SIOPCB(exinf);

	if (sio_putready(siopcb)) {
		sio_irdy_snd(siopcb->exinf);
	}
	if (sio_getready(siopcb)) {
		sio_irdy_rcv(siopcb->exinf);
	}
}

/*
 *  1M
 */
bool_t sio_snd_chr(SIOPCB *siopcb, char c)
{
	bool_t stat;

	if (sio_putready(siopcb)) {
		sil_wrb_mem((void*)UART_D(siopcb->reg), c);
		stat = true;
	}
	else
	{
		stat = false;
	}

	return stat;
}

/*
 *  1M
 */
int_t sio_rcv_chr(SIOPCB *siopcb)
{
	int_t c = -1;

	if (sio_getready(siopcb)) {
		c = sil_reb_mem((void*)UART_D(siopcb->reg));
	}

	return c;
}

/*
 *  R[obN̋
 */
void sio_ena_cbr(SIOPCB *siopcb, uint_t cbrtn)
{
	switch (cbrtn) {
	case SIO_RDY_SND:
		sil_orb((void*)UART_C2(siopcb->reg), C2_TCIE);
		break;
	case SIO_RDY_RCV:
		sil_orb((void*)UART_C2(siopcb->reg), C2_RIE);
		break;
	default:
		break;
	}
}

/* 
 *  R[obN̋֎~
 */
void sio_dis_cbr(SIOPCB *siopcb, uint_t cbrtn)
{
	switch (cbrtn) {
	case SIO_RDY_SND:
		sil_andb((void*)UART_C2(siopcb->reg), ~C2_TCIE);
		break;
	case SIO_RDY_RCV:
		sil_andb((void*)UART_C2(siopcb->reg), ~C2_RIE);
		break;
	default:
		break;
	}
}

/*
 *  1ói|[Oł̏ój
 */
void sio_pol_snd_chr(char_t c, ID siopid)
{
	uint32_t reg = sioreg_table[INDEX_PORT(siopid)];

	sil_wrb_mem((void*)UART_D(reg), c);

	while ((sil_reb_mem((void*)UART_S1(reg)) & S1_TC) == 0) ;
}
